@@ -12,14 +12,26 @@
 #include <string.h>
 #include <unistd.h>
 
-static void	 child(int, int);
-static void	 parent(int, int, const char *);
-static char	*parsekeys(const char *);
-static void	 sighandler(int);
-static void	 usage(void);
+#include "compat.h"
 
-static char	**pickargv;
-static int	  gotsig;
+__dead static void	 child(int, int);
+static void		 parent(int, int, const char *);
+static char		*parsekeys(const char *);
+static void		 sighandler(int);
+__dead static void	 usage(void);
+
+static char		**pickargv;
+/*
+ * Mandatory environment variables required by pick to operate correctly.
+ * Any existing value will be overwritten.
+ */
+static const char	 *pickenv[] = {
+	"LC_CTYPE",		"en_US.UTF-8",
+	"MALLOC_OPTIONS",	"S",		/* malloc.conf(5) options on OpenBSD */
+	"TERM",			"xterm",
+	NULL,
+};
+static int		  gotsig;
 
 int
 main(int argc, char *argv[])
@@ -42,7 +54,7 @@ main(int argc, char *argv[])
 	/* Ensure room for program and null terminator. */
 	if ((pickargv = calloc(argc + 2, sizeof(const char **))) == NULL)
 		err(1, NULL);
-	pickargv[0] = "pick";
+	pickargv[0] = "./pick";
 	for (i = 0; i < argc; i++)
 		pickargv[i + 1] = argv[i];
 
@@ -77,46 +89,48 @@ main(int argc, char *argv[])
 	return 0;
 }
 
-static void
+__dead static void
 usage(void)
 {
-	fprintf(stderr, "usage: pick-test [-k keys] [-- argument ...]\n");
+	fprintf(stderr, "usage: pick-test [-k path] [-- argument ...]\n");
 	exit(1);
 }
 
 static char *
-parsekeys(const char *s)
+parsekeys(const char *path)
 {
+	FILE	*fh;
 	char	*buf;
 	size_t	 len = 0;
 	size_t	 size = 16;
-	int	 c;
+	int	 c, esc;
+
+	if ((fh = fopen(path, "r")) == NULL)
+		err(1, "fopen: %s", path);
 
 	if ((buf = malloc(size)) == NULL)
 		err(1, NULL);
 
-	for (; (c = *s) != '\0'; s++) {
+	esc = 0;
+	while ((c = fgetc(fh)) != EOF) {
 		if (c == '\\') {
-			switch (*++s) {
-			case 'n':
-				c = '\n';
-				break;
-			default:
-				c = *s;
-			}
-		} else if (c == ' ') {
+			esc = 1;
+		} else if (!esc && c == ' ') {
 			continue;
+		} else {
+			buf[len++] = c;
+			esc = 0;
 		}
-		buf[len++] = c;
 
 		if (size <= len) {
-			if (size > ULONG_MAX/2)
-				errx(1, "buffer size overflow");
-			if ((buf = realloc(buf, 2*size)) == NULL)
+			if ((buf = reallocarray(buf, 2, size)) == NULL)
 				err(1, NULL);
 			size *= 2;
 		}
 	}
+	if (ferror(fh))
+		err(1, "fgetc: %s", path);
+	fclose(fh);
 	buf[len] = '\0';
 
 	return buf;
@@ -128,11 +142,12 @@ sighandler(int sig)
 	gotsig = sig == SIGCHLD;
 }
 
-static void
+__dead static void
 child(int master, int slave)
 {
-	struct winsize	ws;
-	int		fd;
+	const char	**env;
+	struct winsize	  ws;
+	int		  fd;
 
 	close(master);
 
@@ -155,20 +170,17 @@ child(int master, int slave)
 	 * Set window size to known dimensions, necessary in order to deduce
 	 * when scrolling should occur.
 	 */
+	memset(&ws, 0, sizeof(ws));
 	ws.ws_col = 80, ws.ws_row = 24;
 	if (ioctl(slave, TIOCSWINSZ, &ws) == -1)
 		err(1, "TIOCSWINSZ");
 
-	/*
-	 * Enable malloc.conf(5) options on OpenBSD which will abort the pick
-	 * process when reading/writing out-of-bounds or accessing already freed
-	 * memory.
-	 */
-	if (setenv("MALLOC_OPTIONS", "S", 0) == -1)
-		err(1, "setenv");
+	for (env = pickenv; *env != NULL; env += 2)
+		if (setenv(env[0], env[1], 1) == -1)
+			err(1, "setenv: %s", env[0]);
 
-	execvp(pickargv[0], pickargv);
-	err(126 + (errno == ENOENT), "%s", pickargv[0]);
+	execv(pickargv[0], pickargv);
+	err(1, "%s", pickargv[0]);
 }
 
 static void
